#!/usr/bin/perl
use strict;

# myth_SDreport.pl : An HTML report on Recording Storage Group Directories.
#
# Jerome Yuzyk, http://mythtv.HeIsJerome.com
#
# 2016.01.20 - first public release
# 2016.04.13 - added Awaiting Expiry table; some reorg
# 2016.05.18 - added All Storage Groups graph
#
# Uses the Myth Services API to query the master backend for recordings in
# the 'Default' Storage Group, then prepares a non-CSS HTML report with a 
# list and some summaries of recordings for each Storage Directory.
#
# Must be run from the Master Backend to inspect each Storage Directory 
# through system calls because there's no Myth way to determine which SD
# a recording is in.
#
# The backend need not be quiet. Could be put in a cron job.
# Works in 0.27 and 0.28, doesn't used any changed API calls.
#
# Limit of 99999 (%5.0f) files and 99999.99 (%5.2f) GiB for fields


# Customizations vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
#
our $backend = 'localhost';	# moot because must be run from MBE anyway
#
# Note: This file will need to be permitted to your mythtv user and Apache
# in this directory, or this script run as root. Then it can be called as
#                      http://Backend/SDreport.html
#
my $outfile = '/var/www/html/SDreport.html';
#
# Customizations ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


# These modules, or similar, are required to call the Service APIs and
# parse their returns.
#
use LWP::Simple;
use XML::Simple;

my $GiB 	= 1024 ** 3;	# 1 GiB constant
my $TimeStamp	= &getTimeStamp;


# Set up some HTML constants and variables
#
my $pageTitle = "Recording Storage Group Directories on [$backend]";

my $pageStart = '';
my $pageBody  = '';
my $pageEnd = '<p><center><i>'
         . '-'x30 . " End of $pageTitle " . '-'x30
	 . '<i></center></p>' 
         . '</body></html>';

# Rows use sprintf as simple-minded macros. These are the formats.
#
my $rowfmtStd = '<tr>'					# standard row
    . '<td valign=top>%s</td>'
    . '<td valign=top>%s</td>'
    . '<td valign=top>%s</td>'
    . '<td align=right valign=top>%s</td>'
    . "</tr>\n";

my $rowfmtShaded = '<tr bgcolor=%-s>'			# shaded row
    . '<td valign=top>%s</td>'
    . '<td valign=top>%s</td>'
    . '<td valign=top>%s</td>'
    . '<td align=right valign=top>%s</td>'
    . "</tr>\n";

my $rowfmtSummary = '<tr>'				# summary row
    . '<td valign=top>%s</td>'
    . '<td valign=top>%s</td>'
    . '<td align=right valign=top><b>%s</b></td>'
    . '<td align=right valign=top><b>%s</b></td>'
    . "</tr>\n";

my $rowfmtExpire = '<tr>'				# Awaiting Expiry
    . '<td valign=top><font size=2>%s</font></td>'
    . '<td valign=top><font size=2>%s</font></td>'
    . '<td valign=top><font size=2>%s</font></td>'
    . '<td align=right valign=top><font size=2>%s</font></td>'
    . "</tr>\n";

my $rowfmtExpireSummary = '<tr>'			# Awaiting Expiry Summary
    . '<td valign=top><font size=2>%s</font></td>'
    . '<td valign=top><font size=2>%s</font></td>'
    . '<td align=right valign=top><font size=2><b>%s</b></font></td>'
    . '<td align=right valign=top><font size=2><b>%s</b></font></td>'
    . "</tr>\n";

my $rowBlank = sprintf( $rowfmtStd, '&nbsp;', '&nbsp;', '&nbsp;', '&nbsp;' );

my $bgOff = '"#FFFFFF"'; my $bgOn  = '"#E0E0E0"';	# shading colours


# Start preparing the page.
#
$pageStart .= "<html><head><title>$pageTitle</title></head><body>\n";
$pageStart .= '<a name="_Top"></a>'
           .  "<p><center><font size=5><b>$pageTitle</b></font>"
           .  "<br>$TimeStamp&nbsp;&nbsp;&nbsp;&nbsp;"
	   .  "&middot; <i>Deleted and Awaiting Expiry</i></center></p>";

# The body of the page - recording listings per Storage Directory
#
$pageBody .= '<table align=center cellpadding=2>';


#--------------------------------------------------------------------------
#
# Start querying and collecting and calculating.
#
# Get the amount of disk reserved that Myth won't use
#
my $KeepSpace = &getKeepSpace;

# Get the Storage Directories in the Default Storage Group
#     - global to be used by a few subs
#
our @SDdirs = &getStorageDirs;

my %SDstats = %{ &getDiskStats };

# Get the files in each Storage Directory.
#
my %SDfiles;

map { $SDfiles{ $_ } = &getStorageDirFiles( $_ ) } @SDdirs;

# Get the list of recordings from the DB, indexed by filename.
#
my %Recorded = %{ &getRecordedList };

# Recordings per Storage Directory
#
my %ShowsBySD; my %DeletedBySD;

foreach my $sd ( sort keys %SDfiles ) {

    my %Programs;

    $DeletedBySD{ $sd } = 0;

    foreach my $file ( @{ $SDfiles{ $sd }} ) {

	( my $Rec = $Recorded{ $file } ) or next;

	my $Title    = $Rec->{Title};
	my $SubTitle = $Rec->{SubTitle};
	my $FileSize = $Rec->{FileSize};
	my $Deleted  = $Rec->{Deleted};

	$Deleted and $DeletedBySD{ $sd } += $FileSize;

	push @{ $Programs{ $Title } }, {

	    SubTitle => $SubTitle, 
	    FileName => $file, FileSize => $FileSize,
	    Deleted => $Deleted,
	    };
        }

    $ShowsBySD{ $sd } = \%Programs;
    }

# for debugging the %SDfiles hash.
#use Data::Dumper;
#$Data::Dumper::Indent = 1;
#print Dumper( %ShowsBySD );
#exit;

# Write listing and collect counts and stats.
#
my %StatsBySD;
my %ToExpireBySD;

my %StatsByTitle_All;

foreach my $sd ( sort keys %SDfiles ) {

    my %StatsByTitle;

    ( my $anchor = $sd ) =~ s'/'_'g;

    # SD tag
    $pageBody .= "\n<tr><td colspan=4>"
        . "<table border='1' bordercolor=$bgOn bgcolor=$bgOn cellpadding='0' cellspacing='0' width='100%'><tr>"
        .     "<td><a name='$anchor'></a><font size=5><b>&nbsp;$sd</b></font></td>"
	.     '<td align=right><a href="#_Top">[top]</a>&nbsp;</td>'
	.     "</tr></table></td></tr>\n";

    $pageBody .= sprintf( $rowfmtStd, 
    
        '<i>Title</i>', '<i>SubTitle</i>', '<i>File</i>', '<i>GiB</i>' );

    # Sorted by Title
    #
    my $last = '';
    my $band = 0;

    my $sdFileSizes  = 0;
    my $sdRecordings = 0;

    # Walk through each Title to prepare a row for each episode and collect
    # summary data.
    #
    foreach my $Title ( sort keys %{ $ShowsBySD{ $sd } } ) {

	my $TitleShown = '&nbsp;';

	# Alternating row colours by Title
	#
	if ( $Title ne $last ) { 
	    $last = $Title; $band = ! $band; $TitleShown = $Title }

	# Collect Episodes of a Series
	#
	my @Episodes = @{ $ShowsBySD{ $sd }->{$Title} };

	my $nEpisodes = @Episodes;
	$sdRecordings += $nEpisodes;

	my $tFileSize        = 0;	# Total files size for this Series
	my $tFileSizeDeleted = 0;	# Total deleted ...

	foreach my $episode ( @Episodes ) {

            my $SubTitle = $episode->{SubTitle};

            my $FileName = $episode->{FileName};
	    my $FileSize = sprintf( "%5.2f", $episode->{FileSize} / $GiB );

	    # Mark Deleted in italics
	    #
	    if ( $episode->{Deleted} ) {

		push @{ $ToExpireBySD{ $sd } },	# save for Awaiting Expiry
		
		    [ $Title, $SubTitle, $FileName, $FileSize ];

	        $SubTitle = "<i>$SubTitle</i>";
	        $FileName = "<i>$FileName</i>";
	        $FileSize = "<i>$FileSize</i>";
		}

            $pageBody .= sprintf( $rowfmtShaded, 
    
	        ( $band ? $bgOn : $bgOff ),	# banding control

	        $TitleShown, $SubTitle, $FileName, $FileSize );

	    $tFileSize += $episode->{FileSize};

	    $episode->{Deleted} and $tFileSizeDeleted += $episode->{FileSize};
            }

	# Entry into %StatsByTitle, used for summary stats
	#
	$StatsByTitle{ $Title } = {
	
	    nEpisodes => $nEpisodes, 
	    tFileSize => $tFileSize, tFileSizeDeleted => $tFileSizeDeleted,
	    };

	$sdFileSizes += $tFileSize;
        }

    $StatsBySD{ $sd }->{Bytes}      = $sdFileSizes;
    $StatsBySD{ $sd }->{Recordings} = $sdRecordings;

    # Total N and GiB
    #
    $pageBody .= sprintf( $rowfmtSummary, '&nbsp;', '&nbsp;', 

	sprintf( "N= %5.0f", $sdRecordings    )
	,
	sprintf( "%5.2f", $sdFileSizes / $GiB ) 
	);

    # List of Awaiting Expiry
    #
    $pageBody .= sprintf( $rowfmtStd, '<B>Awaiting&nbsp;Expiry:</B>', 
    
        '&nbsp;', '&nbsp;', '&nbsp;' );

    my ( $nToExpire, $sumToExpire ) = ( 0, 0 );

    foreach my $e ( @{ $ToExpireBySD{ $sd } } ) {

        $pageBody .= sprintf( $rowfmtExpire, 

	    $e->[0], $e->[1], $e->[2], $e->[3] );

	$nToExpire   += 1;
	$sumToExpire += $e->[3];
        }

    $pageBody .= sprintf( $rowfmtExpireSummary, '&nbsp;', '&nbsp;',
    
	sprintf( "N= %5.0f", $nToExpire   )
	,
	sprintf(    "%5.2f", $sumToExpire )
	);

    # Summary Bar Chart
    #
    foreach my $title ( sort keys %StatsByTitle ) {

	$StatsByTitle{ $title }->{pctOfSD} = sprintf( "%5.2f", 

	    $StatsByTitle{ $title }->{tFileSize} / $sdFileSizes  * 100 );

	$StatsByTitle{ $title }->{pctOfSD_Undeleted} = sprintf( "%5.2f", 

	    ( $StatsByTitle{ $title }->{tFileSize} 
		- $StatsByTitle{ $title }->{tFileSizeDeleted} )

            / $sdFileSizes  * 100 );

	$StatsByTitle{ $title }->{pctOfSD_Deleted} = sprintf( "%5.2f", 

	    $StatsByTitle{ $title }->{tFileSizeDeleted} / $sdFileSizes  * 100 );

	$StatsByTitle{ $title }->{pctGraph} = '|' x 

	    ( int(( $StatsByTitle{ $title }->{pctOfSD_Undeleted} * 4 ) + 0.5 ));

	$StatsByTitle{ $title }->{pctGraphDeleted} = '&middot;' x 

	    ( int(( $StatsByTitle{ $title }->{pctOfSD_Deleted} * 4 ) + 0.5 ));
        }

    $pageBody .= '<tr><td colspan=4><table cellpadding=1>';

    $pageBody .= "<tr><td><i>Title</i></td><td align=right><i>N</i></td><td align=right><i>GiB</i></td><td align=right><i>%</i></td></tr>";

    foreach my $title ( sort keys %StatsByTitle ) {

	my $nEpisodes        = $StatsByTitle{$title}->{nEpisodes};
	my $tFileSize        = sprintf( "%5.2f", 
	    $StatsByTitle{$title}->{tFileSize} / $GiB );
	my $pctOfSD          = $StatsByTitle{$title}->{pctOfSD};
	my $pctGraph         = $StatsByTitle{$title}->{pctGraph};
	my $pctGraphDeleted  = $StatsByTitle{$title}->{pctGraphDeleted};

        $pageBody .= '<tr>'
	    . "<td valign=top>$title</td>"
	    . "<td align=right valign=top>$nEpisodes</td>"
	    . "<td align=right valign=top>$tFileSize</td>"
	    . "<td align=right valign=top>$pctOfSD</td>"
	    . "<td valign=top><nobr>$pctGraph$pctGraphDeleted</nobr></td>"
	    . "</tr>";

	$StatsByTitle_All{ $title }->{nEpisodes} =
	    ( $StatsByTitle_All{ $title }->{nEpisodes} || 0 )
	    + $nEpisodes;

	$StatsByTitle_All{ $title }->{tFileSize} =
	    ( $StatsByTitle_All{ $title }->{tFileSize} || 0 )
	    + $StatsByTitle{$title}->{tFileSize};

	$StatsByTitle_All{ $title }->{tFileSizeDeleted} =
	    ( $StatsByTitle_All{ $title }->{tFileSizeDeleted} || 0 )
	    + $StatsByTitle{$title}->{tFileSizeDeleted};
	}

    $pageBody .= '<tr><td>&nbsp;</td></tr>';

    $pageBody .= '</td></tr></table>';
    }

$pageBody .= '</table>';


# Summary Heading
#
my $summary = '';

$summary .= '<table align=center>';

$summary .= '<tr><td><i>Storage Directory</i></td>'
         .  '<td align=right><i>N</i></td>'
	 .  '<td align=right><i>GiB</i></td>'
	 .  '<td align=right><i>Avg</i></td>'
	 .  '<td align=right colspan=2><i>&middot; Deleted</i></td>'
	 .  '<td align=right><i>Free</i></td>'
	 .  '</tr>';

my ( $t_Recordings, $t_Size, $t_SizeUndeleted, $t_SizeDeleted, $t_Free ) =
   (             0,       0,                0,              0,      0 );

foreach my $sd ( sort keys %StatsBySD ) {

    my $recordings = $StatsBySD{ $sd }->{Recordings};

    my $size = sprintf( "%5.2f", $StatsBySD{ $sd }->{Bytes} / $GiB );
    my $sizeUndeleted = sprintf( "%5.2f", 
    
        ( $StatsBySD{ $sd }->{Bytes} - $DeletedBySD{ $sd } ) / $GiB );

    my $avg = sprintf( "%5.2f", $size / $recordings );

    my $bars = "|" x int(( $sizeUndeleted / 10 ) + 0.5 );

    my $sizeDeleted = sprintf( "%5.2f", ( $DeletedBySD{ $sd } ) / $GiB );
    my $barsDeleted = '&middot;' x int(( $sizeDeleted / 10 ) + 0.5 );

    ( my $anchor = $sd ) =~ s'/'_'g;

    # = Free - KeepSpace
    my $sdFree = sprintf( "%5.2f", 
    
        ( $SDstats{ $sd }->{Free} - $KeepSpace ) / $GiB );

    # Warn if <5% space free in what Myth is allowed to use. 
    #
    my $sdWarning = $size < ( ( $SDstats{ $sd }->{Size} - $KeepSpace ) * 0.95 )
    
        ? '&nbsp;' : '!!!&nbsp;&lt;5%&nbsp;!!!';

    $summary .= "<tr><td><a href='#$anchor'>$sd</a></td>"
             .  "<td align=right>$recordings</td>"
	     .  "<td align=right>$size</td>"
	     .  "<td align=right>$avg</td>"
	     .  "<td>$bars$barsDeleted</td>"
	     .  "<td align=right>$sizeDeleted</td>"
	     .  "<td align=right>$sdFree</td>"
	     .  "<td align=center>$sdWarning</td>"
	     .  "</tr>";

    $t_Recordings	+= $recordings;
    $t_Size		+= $StatsBySD{ $sd }->{Bytes};
    $t_SizeUndeleted	+= $StatsBySD{ $sd }->{Bytes} - $DeletedBySD{ $sd };
    $t_SizeDeleted 	+= $DeletedBySD{ $sd };
    $t_Free	 	+= $SDstats{ $sd }->{Free} - $KeepSpace;
    }

# Totals

my $t_Avg = sprintf( "%5.2f", $t_Size / $t_Recordings / $GiB );

my ( $tSize_Raw, $tSizeUndeleted_Raw, $tSizeDeleted_Raw ) =
   ( $t_Size,    $t_SizeUndeleted,    $t_SizeDeleted,   );

$t_Size			= sprintf( "%5.2f", $t_Size          / $GiB );
$t_SizeUndeleted	= sprintf( "%5.2f", $t_SizeUndeleted / $GiB );
$t_SizeDeleted		= sprintf( "%5.2f", $t_SizeDeleted   / $GiB );
$t_Free			= sprintf( "%5.2f", $t_Free          / $GiB );

$summary .= "<tr><td><a href='#AllGroups'>All&nbsp;Groups</a></td>"
    . "<td align=right>$t_Recordings</td>"
    . "<td align=right>$t_Size</td>"
    . "<td align=right>$t_Avg</td>"
    . "<td            >$t_SizeUndeleted</td>"
    . "<td align=right>$t_SizeDeleted</td>"
    . "<td align=right>$t_Free</td>"
    . "</tr>";

$summary .= '</table>';

# Graph of all Titles
#
my $alltitlesgraph = '<table width=100%>'

    . "\n<tr><td colspan=5>"

    . "<table border='1' bordercolor=$bgOn bgcolor=$bgOn cellpadding='0' cellspacing='0' width='100%'><tr>"
    .     "<td><a name='AllGroups'></a><font size=5><b>&nbsp;All Storage Groups</b></font></td>"
    .     '<td align=right><a href="#_Top">[top]</a>&nbsp;</td>'
    .     "</tr></table></td></tr>\n"

    . "<tr><td><i>Title</i></td><td align=right><i>N</i></td><td align=right><i>GiB</i></td><td align=right><i>%</i></td></tr>\n";

foreach my $title ( sort keys %StatsByTitle_All ) {

    my $pct = sprintf( "%5.2f", 

	$StatsByTitle_All{ $title }->{tFileSize} / $tSize_Raw  * 100 );

    my $pctUndeleted = sprintf( "%5.2f", 

	( $StatsByTitle_All{ $title }->{tFileSize} 
	    - $StatsByTitle_All{ $title }->{tFileSizeDeleted} )

        / $tSize_Raw  * 100 );

    my $pctDeleted = sprintf( "%5.2f", 

	( $StatsByTitle_All{ $title }->{tFileSizeDeleted} 

        / $tSize_Raw  * 100 ));

    my $nEpisodes        = $StatsByTitle_All{$title}->{nEpisodes};
    my $tFileSize        = sprintf( "%5.2f", 
	$StatsByTitle_All{$title}->{tFileSize} / $GiB );
    my $pctGraph         = '|'        x ( int(( $pctUndeleted * 5 ) + 0.5 ));
    my $pctGraphDeleted  = '&middot;' x ( int(( $pctDeleted   * 5 ) + 0.5 ));

    $alltitlesgraph .= '<tr>'
	    . "<td valign=top>$title</td>"
	    . "<td align=right valign=top>$nEpisodes</td>"
	    . "<td align=right valign=top>$tFileSize</td>"
	    . "<td align=right valign=top>$pct</td>"
	    . "<td valign=top><nobr>$pctGraph$pctGraphDeleted</nobr></td>"
	    . "</tr>";
    }

$alltitlesgraph .= '</table>';

# Write everything
#
open OUT, ">$outfile";

print OUT $pageStart;
print OUT $summary;
print OUT $pageBody;
print OUT $alltitlesgraph;
print OUT $pageEnd;

close OUT;	# ALL DONE - go read


#--  Support Routines  ---------------------------------------------------

sub getMythVersion
{
    # One way to get the Myth Version (e.g., 0.27.20151025-1)
    #
    # http://www.gossamer-threads.com/lists/mythtv/dev/592985#592985

    my $query = 'Channel/GetVideoSourceList';

    my $xmlstring = get "http://${backend}:6544/${query}" 
        || return 0;	#==> what should it be?

    my $xml = new XML::Simple;
    my $parsedxml = $xml->XMLin( $xmlstring );

    return $parsedxml->{Version};
    }

sub isMythVersion
{
    # Does the reported version match ^requested
    #
    my ( $reqVersion ) = @_;

    my $haveVersion = &getMythVersion || '?';

    return $haveVersion =~ m/^$reqVersion/ ? 1 : 0;
    }

sub getKeepSpace
{
    # Get the amount of disk that Myth won't use
    #
    # https://www.mythtv.org/wiki/Myth_Service#GetSetting

    my $query = 'Myth/GetSetting';

    my $xmlstring = get "http://${backend}:6544/${query}" 
        || return -1;	#==> what should it be?

    my $xml = new XML::Simple;
    my $parsedxml = $xml->XMLin( $xmlstring );

    my @Settings = @{ $parsedxml->{Settings}->{String} };

    my $KeepSpace;

    foreach my $setting ( @Settings ) {

	next unless $setting->{Key} eq 'AutoExpireExtraSpace';

	$KeepSpace = $setting->{Value}; last;
        }

    return $KeepSpace * ( 1024 ** 3 );
    }

sub getRecordedList
{
    # Get the big list of everything that's been recorded and not Expired.
    #
    # https://www.mythtv.org/wiki/DVR_Service#GetRecordedList

    my $query = 'Dvr/GetRecordedList';

    my $xmlstring = get "http://${backend}:6544/${query}" 
        || return ();

    my $xml = new XML::Simple;
    my $parsedxml = $xml->XMLin( $xmlstring );

    my %Recordings;

    foreach my $rec ( @{ $parsedxml->{Programs}->{Program} } ) {

        my $ndx = $rec->{FileName};

	# SubTitle is empty hash if null
	#
	my $subtitle    = $rec->{SubTitle};
	my $description = $rec->{Description};

	# Sometimes SubTitle is not defined. Use Description if it exists,
	# or the recording's StartTs.
	#
	if ( ref( $subtitle ) eq 'HASH' ) {

	    if ( ref( $description ) eq 'HASH' ) {

	        $subtitle = $rec->{Recording}->{StartTs};
		}

	    else {

	        $subtitle =  substr( $description, 0, 40 );
		}
	    }

	$Recordings{ $ndx } = {
	
	    'Title'    => $rec->{Title},
	    'SubTitle' => $subtitle,
	    'FileSize' => $rec->{FileSize},

	    # Might use this to flag Deleted 
	    'Deleted'  => ( $rec->{Recording}->{RecGroup} eq 'Deleted' ) ? 1 : 0,
	    };
        }

    return \%Recordings;
    }

sub getStorageDirs
{
    my $query = 'Myth/GetStorageGroupDirs';

    my $xmlstring = get "http://${backend}:6544/${query}" 
        || return ();

    my $xml = new XML::Simple;
    my $parsedxml = $xml->XMLin( $xmlstring );

    return map{ $_->{DirName} }
    
        @{ $parsedxml->{StorageGroupDirs}->{StorageGroupDir} };
    }

sub getStorageDirFiles
{
    my ( $StorageDir ) = @_;

    my @files = `ls $StorageDir`;
    chomp @files;
    return \@files;
    }

sub getDiskStats
{
    my %DF;

    foreach my $sd ( @SDdirs ) {

	my $sdRoot = '/' . ( split '/', $sd )[1];

        my $DFraw = `df $sdRoot | grep "$sdRoot"`;

	chomp $DFraw;

        $DFraw =~ m/^(\S+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)%\s+(\S+)$/;

        my ( $fs, $size, $used, $free, $upct, $mnt ) = 
	   (  $1,    $2,    $3,    $4,    $5,   $6 );

        $DF{ $sd }->{Size} = $size * 1024;
        $DF{ $sd }->{Used} = $used * 1024;
        $DF{ $sd }->{Free} = $free * 1024;
        }

    return \%DF;
    }

sub getTimeStamp
{
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime( time );

    return sprintf ( "%04d-%02d-%02d %02d:%02d:%02d",
                     $year+1900, $mon+1, $mday, $hour, $min, $sec);
    }

__END__

:r! sudo perl %
